#!/usr/bin/env python
##
##  pyvnc2swf - vnc2swf.py
##
##  $Id: vnc2swf.py,v 1.20 2006/02/09 09:39:18 euske Exp $
##
##  Copyright (C) 2005 by Yusuke Shinyama (yusuke at cs . nyu . edu)
##  All Rights Reserved.
##
##  This is free software; you can redistribute it and/or modify
##  it under the terms of the GNU General Public License as published by
##  the Free Software Foundation; either version 2 of the License, or
##  (at your option) any later version.
##
##  This software is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##  GNU General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this software; if not, write to the Free Software
##  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
##  USA.
##

import sys, os, time, socket, re
import Tkinter, tkFileDialog, tkMessageBox
from tkSimpleDialog import Dialog
from struct import pack, unpack

from movie import SWFInfo
from output import SWFShapeStream, SWFVideoStream
from rfb import RFBError, RFBNetworkClient, RFBNetworkClientForRecording, RFBStreamConverter
stderr = sys.stderr


##  tkPasswordDialog
##
class tkPasswordDialog(Dialog):

  def __init__(self, title, prompt, parent=None):
    if not parent:
      parent = Tkinter._default_root
    self.prompt = prompt
    Dialog.__init__(self, parent, title)
    return
  
  def destroy(self):
    self.entry = None
    Dialog.destroy(self)
    return
  
  def body(self, master):
    w = Tkinter.Label(master, text=self.prompt, justify=Tkinter.LEFT)
    w.grid(row=0, padx=5, sticky=Tkinter.W)
    self.entry = Tkinter.Entry(master, name="entry", show="*")
    self.entry.grid(row=1, padx=5, sticky=Tkinter.W+Tkinter.E)
    return self.entry
  
  def validate(self):
    self.result = self.entry.get()
    return 1


##  RFBNetworkClientWithTkMixin
##
class RFBNetworkClientWithTkMixin:

  def tk_init(self, root):
    self.root = root
    self.doloop = True
    return

  def interrupt(self):
    self.doloop = False
    return

  def loop(self):
    self.doloop = True
    while self.doloop:
      self.root.update()
      if not self.loop1(): break
    self.finish_update()
    return self

  def getpass(self):
    return tkPasswordDialog('Login',
                            'Password for %s:%d' % (self.host, self.port),
                            self.root).result

class RFBNetworkClientWithTk(RFBNetworkClientWithTkMixin, RFBNetworkClient): pass
class RFBNetworkClientForRecordingWithTk(RFBNetworkClientWithTkMixin, RFBNetworkClientForRecording): pass


##  VNC2SWFWithTk
##
class VNC2SWFWithTk:

  TYPES = { 'SWF(Shape)':'shape', 'SWF(Video)':'video', 'VNCRec':'vnc' }

  def __init__(self, info, outtype='shape', host='localhost', port=5900,
               preferred_encoding=(0,), subprocess=None, debug=0):
    self.info = info
    self.debug = debug
    self.preferred_encoding = preferred_encoding
    self.subprocess = subprocess
    self.client = None
    self.root = Tkinter.Tk()
    self.root.wm_protocol('WM_DELETE_WINDOW', self.quit)
    frame1 = Tkinter.Frame(master=self.root)
    frame2 = Tkinter.Frame(master=self.root)
    self.rfbserver = Tkinter.Entry(master=frame1)
    self.rfbserver.insert(0, '%s:%d' % (host, port))
    self.rfbserver.pack(side=Tkinter.LEFT, expand=1)
    self.toggle = Tkinter.Button(master=frame1, state="disabled")
    self.toggle.pack(side=Tkinter.LEFT, expand=1)
    self.savetype = Tkinter.StringVar()
    revtype = dict([ (v,k) for (k,v) in self.TYPES.iteritems() ])
    self.savetype.set(revtype[outtype])
    self.savemenu = apply(Tkinter.OptionMenu, (frame2, self.savetype,)+tuple(self.TYPES.keys()))
    self.savemenu.pack(side=Tkinter.LEFT, expand=1)
    self.saveas = Tkinter.Button(master=frame2, text='Save As...', command=self.set_filename)
    self.saveas.pack(side=Tkinter.LEFT, expand=1)
    frame1.pack(fill=Tkinter.X)
    frame2.pack(fill=Tkinter.X)
    self.doquit = False
    if self.info.filename:
      self.set_filename(self.info.filename)
    return

  def run(self):
    self.status()
    Tkinter.mainloop()
    return

  def quit(self):
    if self.client:
      self.client.interrupt()
      self.doquit = True
    else:
      self.root.destroy()
    return

  def error(self, msg, e):
    print >>stderr, e
    tkMessageBox.showerror('vnc2swf: %s' % msg, str(e))
    return

  def status(self):
    if not self.info.filename:
      self.root.title('vnc2swf: Select file')
      self.toggle.config(text='Start')
      self.toggle.config(state="disabled")
    elif not self.client:
      self.root.title('vnc2swf: Ready')
      self.toggle.config(text='Start')
      self.toggle.config(state="normal")
      self.toggle.config(background='#80ff80')
      self.toggle.config(activebackground='#00ff00')
      self.toggle.config(command=self.record)
    else:
      self.root.title('vnc2swf: Recording')
      self.toggle.config(text='Stop')
      self.toggle.config(state="normal")
      self.toggle.config(background='#ff8080')
      self.toggle.config(activebackground='#ff0000')
      self.toggle.config(command=self.client.interrupt)
    return
  
  def set_filename(self, filename=None):
    outtype = self.TYPES[self.savetype.get()]
    if outtype in ('shape','video'):
      filename = filename or tkFileDialog.asksaveasfilename(
        master=self.root, title='vnc2swf saveas', defaultextension=".swf",
        filetypes=[("Macromedia Flash Files", "*.swf"), ("All Files", "*")])
      if filename:
        self.info.filename = filename
        self.saveas.config(text=os.path.basename(filename))
    elif outtype == 'vnc':
      filename = filename or tkFileDialog.asksaveasfilename(
        master=self.root, title='vnc2swf saveas', defaultextension=".vnc",
        filetypes=[("VNCRec Files", "*.vnc"), ("All Files", "*")])
      if filename:
        self.info.filename = filename
        self.saveas.config(text=os.path.basename(filename))
    self.status()
    self.root.update()
    return
  
  def record(self):
    s = self.rfbserver.get()
    m = re.match(r'^([^:/]*)(:(\d+))?', s)
    if not m:
      tkMessageBox.showerror('vnc2swf: error', 'Invalid address: %r' % s)
      return
    (host, port) = (m.group(1) or 'localhost', int(m.group(3) or '5900'))
    assert self.info.filename
    outtype = self.TYPES[self.savetype.get()]
    if outtype == 'vnc':
      self.client = RFBNetworkClientForRecordingWithTk(host, port, self.info.filename,
                                                       preferred_encoding=self.preferred_encoding)
    else:
      assert outtype in ('shape', 'video')
      if outtype == 'shape':
        stream = SWFShapeStream(self.info)
      else:
        stream = SWFVideoStream(self.info)
      self.client = RFBNetworkClientWithTk(host, port, RFBStreamConverter(self.info, stream),
                                           preferred_encoding=self.preferred_encoding)
    self.client.tk_init(self.root)
    try:
      self.client.init().auth().start()
      if self.debug:
        print 'start recording'
      if self.subprocess:
        self.subprocess.start()
      try:
        self.status()
        self.client.loop()
      finally:
        if self.debug:
          print 'stop recording'
        if self.subprocess:
          self.subprocess.stop()
        self.client.close()
        self.client = None
        self.status()
    except socket.error, e:
      return self.error('Socket error', e)
    except RFBError, e:
      return self.error('RFB protocol error', e)
    if self.doquit:
      self.root.destroy()
    return


def vnc2swf(info, outtype='shape', host='localhost', port=5900,
            preferred_encoding=(0,), subprocess=None, pwdfile=None,
            debug=0):
  if outtype in ('shape','video'):
    if outtype == 'shape':
      stream = SWFShapeStream(info)
    else:
      stream = SWFVideoStream(info)
    converter = RFBStreamConverter(info, stream)
    client = RFBNetworkClient(host, port, converter, pwdfile=pwdfile,
                              preferred_encoding=preferred_encoding)
  else:
    client = RFBNetworkClientForRecording(host, port, info.filename, pwdfile=pwdfile,
                                          preferred_encoding=preferred_encoding)
  try:
    client.init().auth().start()
    if debug:
      print 'start recording'
    if subprocess:
      subprocess.start()
    try:
      try:
        client.loop()
      except KeyboardInterrupt:
        pass
    finally:
      if debug:
        print 'stop recording'
      if subprocess:
        subprocess.stop()
      client.close()
  except socket.error, e:
    print >>stderr, 'Socket error:', e
  except RFBError, e:
    print >>stderr, 'RFB error:', e
  return


# Subprocess management
class Subprocess:
  
  def __init__(self, s):
    try:
      import subprocess
    except ImportError:
      print >>stderr, '-S option requires Python 2.4 or newer.'
      sys.exit(111)
    if not hasattr(os, 'kill'):
      print >>stderr, '-S option works only on Unix or Mac OS X.'
      sys.exit(111)
    self.args = s.split(' ')
    self.popen = None
    return
  
  def start(self):
    import subprocess
    self.popen = subprocess.Popen(self.args)
    return
  
  def stop(self):
    import signal
    os.kill(self.popen.pid, signal.SIGINT)
    self.popen.wait()
    return


# main
# ./vnc2swf.py -S 'arecord -t wav -c 1 -r 22050 out.wav' -n -o out.swf 
def main(args):
  import getopt
  def usage():
    print ("usage: vnc2swf.py [-d] [-n] [-o filename] [-t {shape|video|vnc}]" \
           " [-e encoding] [-N] [-C clipping] [-r framerate] [-S subprocess] [-P pwdfile] [host[:display] [port]]")
    sys.exit(2)
  try:
    (opts, args) = getopt.getopt(sys.argv[1:], 'dno:t:e:NC:r:S:P:')
  except getopt.GetoptError:
    usage()
  (debug, console, outtype, subprocess, pwdfile) = (0, False, None, None, None)
  (cursor, host, port, preferred_encoding) = (True, 'localhost', 5900, (0,))
  info = SWFInfo()
  for (k, v) in opts:
    if k == '-d': debug += 1
    elif k == '-n': console = True
    elif k == '-t': outtype = v
    elif k == '-e': preferred_encoding = tuple([ int(i) for i in v.split(',') ])
    elif k == '-N': cursor = False
    elif k == '-S': subprocess = Subprocess(v)
    elif k == '-P': pwdfile = v
    elif k == '-o':
      info.filename = v
    elif k == '-C':
      m = re.match(r'^(\d+)x(\d+)\+(\d+)\+(\d+)$', v)
      if not m:
        print >>stderr, 'Invalid clipping specification:', v
        usage()
      x = map(int, m.groups())
      info.clipping = (x[2],x[3], x[0],x[1])
    elif k == "-r":
      info.framerate = int(v)
  if not outtype:
    if info.filename:
      if info.filename.endswith('.vnc'):
        outtype = 'vnc'
      elif info.filename.endswith('.swf'):
        outtype = 'shape'
    else:
      outtype = 'shape'
  if outtype not in ('shape','video','vnc'):
    print 'Please specify the output type or file extension.'
    usage()
  if cursor:
    preferred_encoding += (-232,-239,)
  if 1 <= len(args):
    if ':' in args[0]:
      i = args[0].index(':')
      host = args[0][:i] or 'localhost'
      port = int(args[0][i+1:])+5900
    else:
      host = args[0]
  if 2 <= len(args):
    port = int(args[1])
  if console:
    if not info.filename:
      print 'Please specify the output filename.'
      usage()
    vnc2swf(info, outtype, host, port,
            preferred_encoding=preferred_encoding,
            subprocess=subprocess, pwdfile=pwdfile,
            debug=debug)
  else:
    VNC2SWFWithTk(info, outtype, host, port,
                  preferred_encoding=preferred_encoding,
                  subprocess=subprocess,
                  debug=debug).run()

if __name__ == "__main__": main(sys.argv[1:])
